%% Class for dealing with the FFT of a full transmission spectrum with periodicity (e.g. MZI, resonator, ...)
    % overall this class is great for looking at the FSRs and crossings between multiple mode families within a given spectrum
classdef FFT_spectrum
properties 
    lambda_raw  % 1xN wavelength axis - Precondition - has odd number of pts. If even number, then the last point is removed
    T_raw       % 1xN Transmission data points. It is assumed that we are looking for resonance dips (not peaks), so if the raw data has peaks multiply by -1
    N           % number of pts
    name
    lambda      % 1xN resampled wavelength pts (evenly spaced in freq domain)
    w           % 1xN  resampled freq pts
    T           % 1xN real resampled transmission pts
    T_FT        % 1x N/2 complex FFT of transmission. Since fft of real fct is symmetric, we only store and use the single sided FFT
    t_ds        % 1xN double sided time domain axis
    t           % 1x N/2 time axis - the single sided fft of the frequency axis.
    w_range     % bandwidth in freq domain
    t_range     % total time duration in t domain
    dw          % resolution in freq domain
    dt          % resolution in t domain
end
methods 
    % class constructor - just store spectrum and fft of data
    function obj = FFT_spectrum(lambda, T, varargin)
        if nargin > 0 % allow code to not error if no inputs given --> can make empty object arrays
            % If add optional argument "dips_or_peaks", then this sets whether to search for dips in spectrum (default) or peaks
            % Parse inputs.
            p = inputParser;
            pr = @(varargin) p.addParameter(varargin{:});
            pr('dips_or_peaks',"dips",@isstring);       
            pr('name',"mode_spectrum",@isstring);       
            p.parse(varargin{:});
            P = p.Results;
            obj.name = P.name;
            if mod(lambda,2) == 0 % if even number of pts: remove one to get odd length
                obj.lambda_raw = lambda(1:end-1);
                obj.T_raw = T(1:end-1);               
            else 
                obj.lambda_raw = lambda;
                obj.T_raw = T;
            end
            obj.N = length(obj.T_raw); % store number of pts
            % resample and store resampled T and axes in object
            obj = obj.resample();
            % now take FFT
            obj = obj.calc_FFT();
        end
    end
    
    function [obj] = resample(obj)
        w_raw = Helpers.lambdatow(obj.lambda_raw);
        N = length(obj.lambda_raw);
        obj.w = linspace( w_raw(1), w_raw(end), N ); 
        obj.lambda = Helpers.wtolambda(obj.w);
        % calculate and store useful values - resolution and range
        obj.w_range = max(obj.w) - min(obj.w);       obj.dw = abs(obj.w(2) - obj.w(1));
        obj.dt = 2*pi/ obj.w_range;             t_range_ds = 2*pi/obj.dw; obj.t_range = 2*pi/(2*obj.dw); % nyquist      
        
        % start with double sided PSD and then convert to single sided
        obj.t_ds = linspace(-obj.t_range, obj.t_range, N) ;              
        obj.t = obj.t_ds(obj.t_ds >= 0);   

        % resample transmission w interpolation
        obj.T = spline(w_raw', obj.T_raw, obj.w);
                            
    end
    % calculates the FT of an objects spectrum and stores in the object
    function obj = calc_FFT(obj)
        obj.T_FT = FFT_spectrum.FFT( obj.T);
    end
    function obj_LP = calc_LP(obj, t_max)
        % simple lowpass filter multiplies all higher times by 0
        % can be used either to extract background or to smooth data
        LP_kernel = obj.t < t_max;       
        T_FT_LP = obj.T_FT .* LP_kernel;
        T_LP = FFT_spectrum.iFFT(T_FT_LP);
        % Now create a new FFT_spectrum object for the low-passed data
        obj_LP = FFT_spectrum(obj.lambda, T_LP, "name", obj.name + "_LP");
    end
    function obj_pool = calc_max_pool(obj, f_pool)
        T_pool = 0*obj.T;
        % idx_range = freq_range / (dw/dpt)
        idx_range_filter = 2*round(f_pool/2 / (obj.dw/(2*pi)) ); % number of data points to include in each max call, chosen to be an even number
        for i = 1:length(obj.T)
            idx_max = min(length(obj.T), i + idx_range_filter/2);
            idx_min = max(1, i - idx_range_filter/2);
            T_pool(i) = max( obj.T(idx_min : idx_max) );
        end
        obj_pool = FFT_spectrum(obj.lambda, T_pool, "name", obj.name + "_max_pool");        
    end
    function [obj_norm, obj_bg] = normalize_transmission(obj, filter_type, bg_method,  varargin)
        % normalize to unity transmission using method filter_type (either "inverse" or "wiener") and bg_method (either "LP" or "max_pool")
        % if local max pool: t_filter represents the width of data to pool together. If div by low-pass: t_filter represents the t_max to get the low pass data
        p = inputParser;
        p.addParameter("noise", 0); % if use Wiener, then also input the SNR. SNR usually vector: Power (vector) / electronic noise variance (scalar)
        p.addParameter("t_wiener", 0); % s - if use Wiener decide what time to threshold in the low pass to extract the bg
        p.addParameter("f_pool", 0); % Hz - if use max pool, what freq window width to take max and pool together
        p.parse(varargin{:});
        P = p.Results;

        switch bg_method
            case "max_pool"
                obj_bg = calc_max_pool(obj, P.f_pool);
            case "LP"
                obj_bg = calc_LP(obj, P.t_wiener);
            otherwise
                error("invalid bg_method")
        end
        switch filter_type
            case "inverse"
                T_norm = obj.T ./ obj_bg.T;
            case "wiener"
                SNR = obj_bg.T / P.noise;
                SNR_term = 1 ./(1 + 1./( ( obj_bg.T).^2 .* SNR));
                T_norm = obj.T ./  obj_bg.T .* SNR_term;
            otherwise
                error("invalid filter_type")
        end       
        obj_norm = FFT_spectrum(obj.lambda, T_norm, "name", obj.name + "_norm");

    end
%% Wavelet functions

    % function generates an FFT_spectrum object for the wavelet basis function - a narrow lorentzian dip on a broad lorentzian peak such that the net area is 0
    function obj_wavelet = make_Lor_N_B_wavelet(obj, Q_narrow, Q_broad)
        w_det = linspace(-obj.w_range/2, obj.w_range/2, obj.N);
        w_0 = mean(obj.w);
        linewidth_narrow = w_0/Q_narrow;
        linewidth_broad = w_0/Q_broad;
        % generate lorentzians in wavelength domain, then subtract them to get 0 net area
        lor_0 = lorentzian_res(linewidth_narrow, w_det, "amp_or_power", "power", "norm_peak_or_area", "area");        
        lor_1 = lorentzian_res(linewidth_broad, w_det, "amp_or_power", "power", "norm_peak_or_area", "area");
    
        wavelet_T = fftshift(lor_1 - lor_0); % fftshift needed so that conv doesnt shift anything - ok since do cyclic FFT
        % lastly scale to have peak near 1 so get normal range of transmission values (rather than ~ 1e-10)
        wavelet_T = wavelet_T*1e9; % can't divide relative to peak need hardocde, since otherwise get wavelet-dependent scaling
        % now take fft of wavelet
        obj_wavelet = FFT_spectrum(obj.lambda, wavelet_T, "dips_or_peaks", "dips", "name", "Wavelet");
    end
    function obj_wavelet = make_Lor_odd_wavelet(obj, Q, shift_factor)
        w_det = linspace(-obj.w_range/2, obj.w_range/2, obj.N);
        w_0 = mean(obj.w);
        linewidth = w_0/Q;
        % generate lorentzians in wavelength domain, then subtract them to get 0 net area
        lor_0 = lorentzian_res(linewidth, w_det, "amp_or_power", "power", "norm_peak_or_area", "area");        
        lor_1 = lorentzian_res(linewidth, w_det+linewidth*shift_factor, "amp_or_power", "power", "norm_peak_or_area", "area");
    
        wavelet_T = fftshift(lor_1 - lor_0); % fftshift needed so that conv doesnt shift anything - ok since do cyclic FFT
        % lastly scale to have peak near 1 so get normal range of transmission values (rather than ~ 1e-10)
        wavelet_T = wavelet_T*1e9; % can't divide relative to peak need hardocde, since otherwise get wavelet-dependent scaling
        % now take fft of wavelet
        obj_wavelet = FFT_spectrum(obj.lambda, wavelet_T, "dips_or_peaks", "dips", "name", "Wavelet");
    end    
    % function convolves the two spectrum objects by multiplying their FT and then converting back to wavelength domain
    function obj_conv = conv_spectra(obj1, obj2)
        new_T = FFT_spectrum.iFFT(obj1.T_FT .* obj2.T_FT);
        obj_conv = FFT_spectrum(obj1.lambda, new_T, "name", "conv of " + obj1.name + obj2.name);
    end
    % function loops through wavelet basis fct parameters (width) and convolves with obj to give a resulting obj array showing the wavelet transform
    function [obj_transform_array, obj_wvlt_array] = wavelet_Transform(obj, varargin)
        p = inputParser;        pr = @(varargin) p.addParameter(varargin{:});
        pr('wavelet_basis', "Lor_N_B");  % "Lor_odd", "Lor_N_B"
        pr('Q', []);  % for all wavelet kinds
        pr('shift_factor', []);  % % for lor_odd 
        pr('Q_broad', []);  % for Lor_N_B


        p.parse(varargin{:});
        P = p.Results;

        for i = 1:length(P.Q)
            if P.wavelet_basis == "Lor_N_B"
                Q_narrow = P.Q(i); Q_broad = P.Q_broad(i);
                % make wavelet instance and then convolve
                obj_wavelet = make_Lor_N_B_wavelet(obj, Q_narrow, Q_broad);
            elseif P.wavelet_basis == "Lor_odd"
                obj_wavelet = make_Lor_odd_wavelet(obj, P.Q(i), P.shift_factor);                
            end
            obj_transform_array(i) = conv_spectra(obj, obj_wavelet);
            obj_wvlt_array(i) = obj_wavelet;
        end
    end

    function filepath = save_spectrum(obj)
        % save the spectrum object to a .mat file for future reference (don't have to fit everything again)
        save(obj.name, "obj")
    end

    %% Plotting functions
    function h = plot_spectrum_param(obj, str_type, varargin)
        p = inputParser;
        pr = @(varargin) p.addParameter(varargin{:});
        % optional arguments for plot type 
        % pr('pt_size', 10, @isnumeric);  
        pr('plot_dB', true, @islogical);  


        p.parse(varargin{:});
        P = p.Results;

        switch(str_type)
            case("transmission")
                plot(obj.lambda/Helpers.nm, obj.T)                
                xlabel("Wavelength (nm)"); ylabel("Transmission")   
            case("FT")
                if ~P.plot_dB
                    plot(obj.t/Helpers.ns, abs(obj.T_FT).^2)                
                    xlabel("t (ns)"); ylabel("Fourier (abs ^2)")   
                else
                    plot(obj.t/Helpers.ns, 10*log10(abs(obj.T_FT).^2) )               
                    xlabel("t (ns)"); ylabel("Fourier (dB)")   
                end
            case("FT phase")
                phi = unwrap( angle(obj.T_FT));
                plot(obj.t/Helpers.ns, phi/pi)                
                xlabel("t (ns)"); ylabel("Fourier Phase (\pi rad)")  
            case("all") % plot transmission, FT, and phase
                subplot(3,1,1);
                plot_spectrum_param(obj, "transmission")
                subplot(3,1,2);
                plot_spectrum_param(obj, "FT")
                subplot(3,1,3)
                plot_spectrum_param(obj, "FT phase");
        end
    end

end

methods (Static)
        % returns the single-sided fourier transform of real valued function T
    function [T_FT] = FFT(T)
        split_pt = ceil( (length(T) + 1)/2);
        T_FT_ds = fftshift(fft(T)); % fftshift swaps left and right halves of ft so that zero freq is in middle (instead of order by positive freqs on left half and then negative freqs on right half
        T_FT = T_FT_ds(split_pt:end);
    end
    % returns the real valued function T given its single sided fourier transform
    function T = iFFT(T_FT)        
        % go from single sided t domain back to freq domain
        T = real( ifft(ifftshift([conj(fliplr(T_FT(2:end))), T_FT])));       
    end

end
end